function [sys,x0,str,ts] = LeftWaveDenoise(t,x,u,flag,scale,delay,thresh)
%	An S-function which implements a hard threshold denoising of signals using
%   an average-interpolation redundent wavelet transform on the half axis.
%   Input: u :    Single valued data at the current time which
%                 is being buffered.
%   Output:       filtered output
%   Parameters:   scale = number of scales of the wavelet transform
%				  delay = delay between input and output. Non zero to lessen
%				  boundary effect.
%				  thresh = the threshold
%	Francois Chaplais, February 14, 2003, chaplais@cas.ensmp.fr
switch flag,
%%%%%%%%%%%%%%%%%%
% Initialization %
%%%%%%%%%%%%%%%%%%
case 0,
	l = (2^(scale-1))*6+delay; % largeur de la fenetre de donnees
	[sys,x0,str,ts] = mdlInitializeSizes(l);
	%%%%%%%%%%  
	% Update %
	%%%%%%%%%%
case 2,                                               
    sys = mdlUpdate(t,x,u);
	%%%%%%%%%%
	% Output %
	%%%%%%%%%%
case 3,                                               
    sys = mdlOutputs(t,x,u,scale,delay,thresh);    
	%%%%%%%%%%%%%
	% Terminate %
	%%%%%%%%%%%%%
case 9,
    sys = mdlTerminate;
	%%%%%%%%%%%%%%%%%%%%
	% Unexpected flags %
	%%%%%%%%%%%%%%%%%%%%
otherwise
    error(['Unhandled flag = ',num2str(flag)]);
end
% end buffer
%
%=============================================================================
% mdlInitializeSizes
% Return the sizes, initial conditions, and sample times for the S-function.
%=============================================================================
%
function [sys,x0,str,ts] = mdlInitializeSizes(W)
if W < 2
	sizes = simsizes;
	sizes.NumContStates  = 0;
	sizes.NumDiscStates  = 0;
	sizes.NumOutputs     = 1;
	sizes.NumInputs      = 1;
	sizes.DirFeedthrough = 1;
	sizes.NumSampleTimes = 1;
	sys = simsizes(sizes);
	x0  = [];
	str = [];
	ts  = [-1 0];
else
	sizes = simsizes;
	sizes.NumContStates  = 0;
	sizes.NumDiscStates  = W-1;
	sizes.NumOutputs     = 1;
	sizes.NumInputs      = 1;
	sizes.DirFeedthrough = 1;
	sizes.NumSampleTimes = 1;
	sys = simsizes(sizes);
	x0  = zeros(W-1,1);
	str = [];
	ts  = [-1 0];
end
% end mdlInitializeSizes
%
%=======================================================================
% mdlUpdate
% Handle discrete state updates, sample time hits, and major time step
% requirements.
%=======================================================================
%
function sys = mdlUpdate(t,x,u)
W = length(x);
if W < 1
	sys = [];
else
	if W > 1
		x(1:W-1) = x(2:W) ;
	end
	x(W) = u;
	sys = x;
end
%end mdlUpdate
%
%=======================================================================
% mdlOutputs
% Return the output vector for the S-function
%=======================================================================
%
function sys = mdlOutputs(t,x,u,scale,delay,thresh)
wsig.s = [u;x]';
wsig.d = length(x);
WT = LeftWaveTransform(wsig,scale); % transformee en ondelettes
WT = HardThreshold(WT,thresh); % seuillage
z = LeftInvWaveTransform(WT,scale); % restitution du signal
sys = z.s(end-delay); % estimation du signal avec un retard "delay"
%end mdlOutputs
%
%=============================================================================
% mdlTerminate
% Perform any end of simulation tasks.
%=============================================================================
%
function sys = mdlTerminate
sys = [];
% end mdlTerminate

%========================================================================
% signal processing routines
%========================================================================

function WT = LeftWaveTransform(u,scale)
% returns wavelet transform of u using filters h and g
% at scale "scale"
% filter initialization
h.s = [1/2 1/2]*sqrt(2);
h.d = 0;
g.s = [-1/16 -1/16 1/2 -1/2 1/16 1/16]*sqrt(2);
g.d = 2;
WT.LoRes = u;
WT.Details = cell(scale,1); % details are stored columnwise
for i = 1:scale
	e = myconv(WT.LoRes,h);
	d = myconv(WT.LoRes,g);
	d = cutright(d);
	e = cutright(e);
	d = bord_g(WT.LoRes,d,i);
	WT.Details{i} = d;
	WT.LoRes = e;
	% insert zeros into filters
	h = zeroinsert(h);
	g = zeroinsert(g);
end

function y = myconv(x,f)
% convolves two signals
% signals are structs with two elements
% d  is the lowest degree W/R to delay (a delay if nonegative)
% s is the signal on its support
y.d = x.d + f.d;
y.s = conv(x.s,f.s);

function y = cutright(x)
y.s = x.s(1:x.d+1);
y.d = length(y.s)-1;

function y = bord_g(x,y,scale)

nx=length(x.s);
ny=length(y.s);
ns = 2^(scale-1);

F0 = [-1/16 -1/16 1/4 1/4 -11/16 5/16]*sqrt(2);
l = ns*6;
u = x.s(nx-l+1:nx);
u = reshape(u,ns,6);
y.s(ny-ns+1:ny) = F0*u';

F1 = [1/24 1/24 -3/8 11/24 -1/6]*sqrt(2);
l = ns*5;
u = x.s(nx-l+1:nx);
u = reshape(u,ns,5);
y.s(ny-2*ns+1:ny-ns) = F1*u';

function y = zeroinsert(x)
n = size(x.s);
if (n(1) == 1)
	y.s = zeros(1,n(2)*2-1);
else
	y.s = zeros(n(1)*2-1,1);
end
y.s(1:2:end) = x.s;
y.d = 2*x.d;

function WT = HardThreshold(WT,thresh)
scale = length(WT.Details);
for i = 1:scale
	sig = WT.Details{i}.s;
	WT.Details{i}.s = sig .* (abs(sig) >= thresh) ; % threshold
end

function s = LeftInvWaveTransform(WT,scale)
% reconstructs signal conjugate filters
rg.s = [-1 1]/sqrt(2);
rg.d = 1;
rh.s = [-1/8 1/8 1 1 1/8 -1/8]/sqrt(2);
rh.d = 3;
for i = 1:scale
	Filters.rh{i} = rh;
	Filters.rg{i} = rg;
	rh = zeroinsert(rh);
	rg = zeroinsert(rg);
end
for i = scale:-1:1
	d = myconv(WT.Details{i},Filters.rg{i});
	e = myconv(WT.LoRes,Filters.rh{i});
	d = cutright(d);
	e = cutright(e);
	d.s = d.s/2;
	e.s = e.s/2;
	e = bord_rh(WT.LoRes,e,i);
	d.s(end-2^(i-1)+1:end) = -d.s(end-2^i+1:end-2^(i-1));
	WT.LoRes = addSignals(e,d);
end
s = WT.LoRes;

function y = bord_rh(x,y,scale)

nx=length(x.s);
ny=length(y.s);
ns = 2^(scale-1);

F0 = [1/16 1/24 -1/4 -3/8 73/48]/sqrt(2);
l = ns*5;
u = x.s(nx-l+1:nx);
u = reshape(u,ns,5);
y.s(ny-ns+1:ny) = F0*u';

F1 = [-1/16 -1/24  1/4 3/8 23/48]/sqrt(2);
l = ns*5;
u = x.s(nx-l+1:nx);
u = reshape(u,ns,5);
y.s(ny-2*ns+1:ny-ns) = F1*u';

F2 = [-1/16 1/24  1/2 5/8 -5/48]/sqrt(2);
l = ns*5;
u = x.s(nx-l+1:nx);
u = reshape(u,ns,5);
y.s(ny-3*ns+1:ny-2*ns) = F2*u';

function y = addSignals(u,v)
% summation of two Laurent series
% manage lowest degree
if u.d < v.d
	y.d = v.d;
	u.s = [zeros(1,v.d - u.d) u.s];
elseif v.d < u.d
	y.d = u.d;
	v.s = [zeros(1,u.d - v.d) v.s];
else
	y.d = u.d;
end
% manage highest degree
if length(u.s) < length(v.s)
	u.s = [u.s zeros(1,length(v.s)-length(u.s))];
elseif length(v.s) < length(u.s)
	v.s = [v.s zeros(1,length(u.s)-length(v.s))];
end
y.s = u.s + v.s;

